/*  Exile - IRC to Relational Database Gateway
 *  Copyright (C) 2002 Lucas Nussbaum <lucas@lucas-nussbaum.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: messages.c 1676 2006-07-22 17:34:07Z sirvulcan $
 */

#include "config.h"
#include "exile.h"
#include "messages.h"
#include "db.h"
#include "send.h"
#include "misc.h"
#include "compat.h"
#include "memory.h"
#include "log.h"
#include "actions.h"
#include "sockutil.h"

extern char *ServerNum;
extern char *ServerName;
extern char *RemoteServer;
extern int LogAmount;
extern char inbuf[];
extern int HidePasswords;
extern time_t start_time;

extern unsigned int nbusers;
extern unsigned int nbusers_max;
extern unsigned int nbchans;
extern unsigned int nbchans_max;
int log_umode[256];
int log_cmode[256];
int eob_sent = 0;

MYSQL *myptr;

/* List of messages is at the bottom of the file. */

/* PING - handled */
static void m_ping(char *source, int ac, char **av)
{
	if (ac < 1)
		return;

	send_cmd(ServerName, "Z %s %s", ac > 1 ? av[1] : ServerName, av[0]);
}

/* SERVER - forwarded */
/* char *server, char *comment, char *linkedto */
/* SERVER [SERVERNAME] [HOPCOUNT] [START TIME] [LINK TIME] [PROTOCOL] [NUMERIC/MAXCONN] :[DESCRIPTION] */
/* SERVER afro4.funkpolice.com  1 1121933417 1123010310 J10 AaAP] +h :Hell IRC Network */
/* Aa   S gnuworld.afternet.org 2 0          1122878038 J10 Az]]] +s :Services */
/*        0                     1 2          3          4   5     6   7        */
static void m_server(char *source, int ac, char **av)
{
	if (*source)
		do_server(av[0], av[7], source, av[5]);
	else
		do_server(av[0], av[7], ServerName, av[5]);
}

static void m_account(char *source, int ac, char **av)
{
        int c = 1;
	char *acc;
	char *useraccount = "";
	char *timestamp = "";
        char *acctb, *acct = NULL;

#if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
	if ((ac > 3) && (!strcmp(av[1], "R")))
		acc = av[2];
	else
		acc = av[1];
#else
	acc = av[1];
#endif


        for (acct = strtok_r(acc, ":", &acctb);
                acct;
                acct = strtok_r(NULL, ":", &acctb)) {
                if (c == 1)
                        useraccount = acct;
                else if (c == 2)
                        timestamp = acct;

                c++;
        }
        c = 1;


	db_query("UPDATE " TBL_USER
		 " SET account=\"%s\" WHERE snumeric=\"%s\"",
		 useraccount, av[0]);

	db_query("UPDATE " TBL_USER
		 " SET acctimestamp=\"%s\" WHERE snumeric=\"%s\"",
		 timestamp, av[0]);

	db_query("UPDATE " TBL_USER
		 " SET mode_lr=\"%s\" WHERE snumeric=\"%s\"",
		 "Y", av[0]);

}

static void m_gline(char *source, int ac, char **av)
{
	int exist = 0;
	int glineid = 0;
	int action = 0; /* 0 = add, 1 = remove */
	int masklen;
	char *lastmod;
	char *reason;
	char *gmask;
        char **res;
        MYSQL_RES *resptr;

	gmask = av[1];
	masklen = strlen(gmask);

	while (*gmask) {
		switch (*gmask) {
			case '+':
				action = 0;
				break;

			case '-':
				action = 1;
				break;

			default:
				break;
		}

		if (masklen - strlen(gmask) == 1)
			break;
		gmask++;
	}

        db_query("SELECT glineid FROM " TBL_GLINE " WHERE gmask=\'%s\'", gmask);

	if (ac == 4)
		lastmod = "0";
	else
		lastmod = av[3];

        resptr = mysql_store_result(myptr);
        while ((res = mysql_fetch_row(resptr))) {
                glineid = atoi(db_escape(res[0]));
                exist = 1;
	}
	mysql_free_result(resptr);

	reason = db_escape(av[ac-1]);

	if ((exist == 0) && (action == 0))
		db_query("INSERT INTO " TBL_GLINE
			 " (gmask, expire, lastmod, reason) VALUES (\'%s\', \'%ld\',\'%s\',\"%s\")",
			 gmask, time(NULL) - atoi(av[2]), lastmod, reason);
	else if ((exist == 1) && (action == 0))
		db_query("UPDATE " TBL_GLINE
			 " SET gmask='%s', expire='%ld', lastmod='%s', reason=\"%s\" WHERE glineid=\"%d\"",
			 gmask, time(NULL) - atoi(av[2]), lastmod, reason, glineid);
	else if ((exist == 1) && (action == 1))
		db_query("DELETE FROM " TBL_GLINE " WHERE glineid='%d'", glineid);
			
	exist = 0;
}

static void m_burst(char *source, int ac, char **av)
{
	int users = 0, chk = 1, be = 0;
	int snum = 0, nickid = 0, chanid;
	int modes = 0, skip = 0, next = 0;
	char *nickb, *nick = NULL;
	char *chanub, *chanu = NULL;
	char *cnick = 0;

	chanid = db_getchancreate(av[0]);

	if (!av[2])
		return;

	if (strchr(av[2], '+'))
		modes = 1;

	if (modes == 1) {
		char *modestr = av[2];
		skip = 0;
		while (*modestr) {
			switch (*modestr) {
				case 'l':
					skip++;
					break;

				case 'k':
					skip++;
					break;

				default:
					break;
			}
			modestr++;
		}

	}


	if (modes == 1)
		next = skip + 3;
	else
		next = skip + 2;


	if (av[next]) {
		if (!strchr(av[next], '%')) {
			users = next;
			next++;

			if (av[next]) {
				if (strchr(av[next], '%'))
					be = next;
			}
		} else {
			users = 0;
			be = next;
		}
	}


	snum = users;

	int opped = 0;
	int hopped = 0;
	int voiced = 0;

	if (users > 0) {
		for (chanu = strtok_r(av[snum], ",", &chanub);
			chanu;
			chanu = strtok_r(NULL, ",", &chanub))
		{
			char *user;
			char *oplevel = 0;

			chk = 1;

			for (nick = strtok_r(chanu, ":", &nickb);
				nick;
				nick = strtok_r(NULL, ":", &nickb))
			{

				if (chk == 1) {
					cnick  = nick;
			                nickid = db_checknumeric(nick);

			                if (nickid != -1) {
#if defined(IRCD_NEFARIOUS4) || defined(IRCD_NEFARIOUS10)
		                	        db_query("INSERT INTO " TBL_ISON
        	                        	         " (nickid, chanid, mode_lo, mode_lh, mode_lv) VALUES (\'%d\', \'%d\',\'%s\',\'%s\',\'%s\')",
                	                        	   nickid, chanid, "N", "N", "N");

#elif defined(IRCD_IRCU12)
			                        db_query("INSERT INTO " TBL_ISON
        		                                 " (nickid, chanid, mode_lo, mode_lv, oplevel) VALUES (\'%d\', \'%d\',\'%s\',\'%s\',\'0\')",
                		                           nickid, chanid, "N", "N");
#else
		                	        db_query("INSERT INTO " TBL_ISON
        	                        	         " (nickid, chanid, mode_lo, mode_lv) VALUES (\'%d\', \'%d\',\'%s\',\'%s\')",
                	                        	   nickid, chanid, "N", "N");
#endif
					} else
			                        mylog("received join of non-existing user %s on channel ID %d",
                		                        nick, av[0]);

					chk = 2;
				} else {
					opped = 0;
					hopped = 0;
					voiced = 0;

					while (*nick) {
						switch (*nick) {
							case 'o':
								opped = 1;
								break;
							case 'h':
								hopped = 1;
								break;
							case 'v':
								voiced = 1;
								break;

							default:
								oplevel = nick;
								break;
						}
						nick++;

					}
				}
			}

			if ((opped == 1) || (hopped == 1) || (voiced == 1)) {
       	                        user = db_escape(cnick);

               	                db_query
                       	                ("UPDATE " TBL_ISON
                               	         " SET mode_lv=\"%s\" WHERE chanid=\"%d\" AND nickid=\"%d\"",
                                       	(voiced ? "Y" : "N"), chanid, nickid);

#if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
                                db_query
       	                                ("UPDATE " TBL_ISON
               	                         " SET mode_lh=\"%s\" WHERE chanid=\"%d\" AND nickid=\"%d\"",
                       	                (hopped ? "Y" : "N"), chanid, nickid);
#endif
                                db_query
       	                                ("UPDATE " TBL_ISON
               	                         " SET mode_lo=\"%s\" WHERE chanid=\"%d\" AND nickid=\"%d\"",
                       	                 (opped ? "Y" : "N"), chanid, nickid);

#if defined(IRCD_IRCU12)
				db_query
					("UPDATE " TBL_ISON
					 " SET oplevel=\"%s\" WHERE chanid=\"%d\" AND nickid=\"%d\"",
					 oplevel, chanid, nickid);
#endif

				chk = 1;
			}
		}
	}

	if (modes == 1) {
	        int atleastone = 0;
	        char db[1000];
	        char tmp[14] = "mode_XX=\"X\", ";
	        char *modes = av[2];
	        int argptr = 3;

	        if (*modes == '0')
	                return;

	        strcpy(db, "UPDATE " TBL_CHAN " SET ");

	        while (*modes) {
	                switch (*modes) {
		                case '+':
        		                tmp[9] = 'Y';
                		        break;
		                case '-':
	        	                tmp[9] = 'N';
	                	        break;
		                default:
	                        	if (!strchr(CHANMODES, *modes)) {
		                                if (!log_cmode[(int) *modes]) {
	        	                                mylog("unknown mode : chanmode %c (in %s)", *modes, inbuf);
	                	                        log_cmode[(int) *modes] = TRUE;
	                        	        }
		                        }
	        	                else if (*modes == 'b' || *modes == 'e' || *modes == 'I')       /* ignore them */
		                                argptr++;

		                        else if (*modes == 'o' || *modes == 'v')
		                        {
		                                char *user;
		                                user = db_escape(av[argptr++]);
		                                db_query
		                                        ("UPDATE " TBL_ISON
                		                         " SET mode_l%c=\"%c\" WHERE chanid=\"%d\" AND nickid=\"%d\"",
		                                         *modes, tmp[9], chanid, db_getnumeric(user));
		                                free(user);
                		        }
		                        else
		                        {
                		                atleastone = 1;
		                                tmp[5] = ((*modes >= 'a') ? 'l' : 'u');
		                                tmp[6] = tolower(*modes);
		                                strcat(db, tmp);
		                                if (*modes == 'k')
		                                {
		                                        if (tmp[9] == 'Y')
		                                        {
		                                                char *key = db_escape(av[argptr++]);
		                                                if (HidePasswords)
		                                                {
		                                                        strcat(db, "mode_lk_data=\"HIDDEN\", ");
		                                                }
		                                                else
		                                                {
		                                                        strcat(db, "mode_lk_data=\"");
		                                                        strcat(db, key);
		                                                        strcat(db, "\", ");
		                                                }
		                                                free(key);
		                                        }
		                                        else
		                                        {
		                                                strcat(db, "mode_lk_data=\"\", ");
		                                                argptr++;         /* mode -k needs a parameter */
		                                        }
		                                }
						#if defined(IRCD_IRCU12)
		                                else if (*modes == 'A')
		                                {
		                                        if (tmp[9] == 'Y')
		                                        {
		                                                char *key = db_escape(av[argptr++]);
		                                                if (HidePasswords)
		                                                {
		                                                        strcat(db, "mode_uA_data=\"HIDDEN\", ");
		                                                }
		                                                else
		                                                {
		                                                        strcat(db, "mode_uA_data=\"");
		                                                        strcat(db, key);
		                                                        strcat(db, "\", ");
		                                                }
		                                                free(key);
		                                        }
		                                        else
		                                        {
		                                                strcat(db, "mode_uA_data=\"\", ");
		                                                argptr++;         /* mode -A needs a parameter */
		                                        }
		                                }
		                                else if (*modes == 'U')
		                                {
		                                        if (tmp[9] == 'Y')
		                                        {
		                                                char *key = db_escape(av[argptr++]);
		                                                if (HidePasswords)
		                                                {
		                                                        strcat(db, "mode_uU_data=\"HIDDEN\", ");
		                                                }
		                                                else
		                                                {
		                                                        strcat(db, "mode_uU_data=\"");
		                                                        strcat(db, key);
		                                                        strcat(db, "\", ");
		                                                }
		                                                free(key);
		                                        }
		                                        else
		                                        {
		                                                strcat(db, "mode_uU_data=\"\", ");
		                                                argptr++;         /* mode -U needs a parameter */
		                                        }
		                                }
						#endif
		                                else if (*modes == 'l')
		                                {
		                                        if (tmp[9] == 'Y')
		                                        {
		                                                strcat(db, "mode_ll_data=\"");
		                                                strcat(db, av[argptr++]);
		                                                strcat(db, "\", ");
		                                        }
		                                        else
		                                        {
		                                                strcat(db, "mode_ll_data=\"\", ");
		                                        }
		                                }
		                        }
		                        break;
		                }
		                modes++;
		        }
#ifndef NOMODES
	        if (atleastone)
	        {
	                sprintf(&db[strlen(db) - 2], " WHERE chanid=\'%d\'", chanid);
	                db_query(db);
	        }
#endif /* !NOMODES */
	}

	if (be > 0) { /* bans and exempts */
                #if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
                int echeck = 0;
                #endif
		int ctype = 0, bcheck = 0;
		char* parm  = NULL;

		for(parm = mysep(&av[ac-1], " "); /* parm = first param */
		    parm;   /* While param is not null */
		    parm = mysep(&av[ac-1], " ")) {  /* parm = next param */
			switch (parm[0]) {
				case '%':
					ctype = 1;
					break;
                                #if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
				case '~':
					ctype = 2;
					break;
                                #endif
		                default:
					break;
			}

			if (ctype == 1) {
				if (bcheck == 0) {
					/* strip % char off start of very first ban */
					if (strlen(parm) > 1) {
			                        db_query("INSERT INTO " TBL_CHBAN
	        	                                 " (chanid, ban) VALUES (\'%d\', \'%s\')",
                		                           chanid, db_escape(strtok(parm, "%")));
					}
					bcheck = 1;
				} else {
		                        db_query("INSERT INTO " TBL_CHBAN
        	                                 " (chanid, ban) VALUES (\'%d\', \'%s\')",
               		                           chanid, db_escape(parm));
				}
                        }
                        #if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
			else if (ctype == 2) {
				if (echeck == 0) {
					echeck = 1;
				} else {
		                        db_query("INSERT INTO " TBL_CHEXEMPT
        	                                 " (chanid, exempt) VALUES (\'%d\', \'%s\')",
               		                           chanid, db_escape(parm));
				}
			}
                        #endif
		}
        }


}

/* SQUIT - forwarded */
static void m_squit(char *source, int ac, char **av)
{
	do_squit(av[0]);
}

/* NICK - forwarded */
static void m_nick(char *source, int ac, char **av)
{
	if (ac > 2)
		do_nick_new(ac, av, source);
	else
		do_nick_chg(av[0], source);
}

/* KICK - forwarded */
static void m_kick(char *source, int ac, char **av)
{
	do_part(av[0], av[1]);		  /* kicks are handled like parts */
}

/* PART - forwarded */
static void m_part(char *source, int ac, char **av)
{
	int chanid;

	chanid = db_checkchannel(av[0]); /* temp fix to a wierd bug */
	if (chanid == -1)
		return;
	else
		do_part(av[0], source);
}

/* ERROR - handled */
static void m_error(char *source, int ac, char **av)
{
	fatal(av[0][0] == ':' ? av[0] + 1 : av[0]);
}

/* JOIN - forwarded */
static void m_join(char *source, int ac, char **av)
{
	char *pos, *oldpos;
	oldpos = av[0];

        if (!strcmp(av[0], "0")) {
                do_partall(source);
		return;
	}

	while ((pos = strchr(oldpos, ',')) != NULL)
	{
		*pos = '\0';
		do_join(oldpos, source, 0);
		oldpos = pos + 1;
	}
	do_join(oldpos, source, 0);
}

static void m_create(char *source, int ac, char **av)
{
	char *pos, *oldpos;
	/* LAME ultimate 2.8 with multiple channels in the same JOIN ... */
	oldpos = av[0];
	while ((pos = strchr(oldpos, ',')) != NULL)
	{
		*pos = '\0';
		do_join(oldpos, source, 1);
		oldpos = pos + 1;
	}
	do_join(oldpos, source, 1);
}

/* QUIT - forwarded */
static void m_quit(char *source, int ac, char **av)
{
	do_quit(source);
}

static void m_clearmode(char *source, int ac, char **av)
{
	do_clearmode(av[0], &av[1], ac - 1);
}

/* MODE - forwarded */
static void m_mode(char *source, int ac, char **av)
{
	if (*av[0] != '#')
		do_umode(av[0], av[1], av[2] ? av[2] : 0);
	else
		do_cmode(av[0], &av[1], ac - 1);
}

/* TOPIC - forwarded */
static void m_topic(char *source, int ac, char **av)
{

#ifdef IRCD_NEFARIOUS10
        char *chan = strdup(av[0]);
	do_topic(chan, av[1], av[2], av[4]);
#else
	char *setter = 0;
        char **res;
        MYSQL_RES *resptr;

	if (strlen(source) > 2) {
		int nickid;
		nickid = db_getnumeric(source);

	        db_query("SELECT nick FROM " TBL_USER " WHERE nickid=\'%d\'", nickid);
	} else {
		int servid;
		servid = db_getserver(source);

	        db_query("SELECT server FROM " TBL_SERV " WHERE servid=\'%d\'", servid);
	}

        resptr = mysql_store_result(myptr);
        while ((res = mysql_fetch_row(resptr)))
                setter = db_escape(res[0]);

	mysql_free_result(resptr);

	do_topic(av[0], setter, av[2], av[3]);
#endif
}

/* KILL - forwarded */
static void m_kill(char *source, int ac, char **av)
{
	do_quit(av[0]);
}

/* AWAY - forwarded */
static void m_away(char *source, int ac, char **av)
{
	if (!ac)
		do_away_unset(source);
	else
		do_away_set(source, av[0]);
}

/* VERSION - handled */
static void m_version(char *source, int ac, char **av)
{
	if (source)
		send_cmd(NULL, "%s 351 %s Exile %s %s",
			ServerNum, source, VERSION, ServerName);
}

/* STATS - handled */
static void m_stats(char *source, int ac, char **av)
{
	if (ac < 1)
		return;
	switch (*av[0])
	{
	case 'l':
		send_cmd(NULL,
					"%s 211 %s Server SendBuf SentBytes SentMsgs RecvBuf RecvBytes RecvMsgs ConnTime",
					 ServerNum, source);
		send_cmd(NULL, "%s 211 %s %s %d %d %d %d %d %d %ld", ServerNum, source,
					RemoteServer, write_buffer_len(), total_written, -1,
					read_buffer_len(), total_read, -1, time(NULL) - start_time);
		send_cmd(NULL, "%s 219 %s l :End of /STATS report.", ServerNum, source);
		break;
	case 'u':
		{
			int uptime;
			uptime = time(NULL) - start_time;
			send_cmd(NULL, "%s 242 %s :Exile up %d day%s, %02d:%02d:%02d",
						ServerNum, source, uptime / 86400, (uptime / 86400 == 1) ? "" : "s",
						(uptime / 3600) % 24, (uptime / 60) % 60, uptime % 60);
			send_cmd(NULL, "%s 219 %s u :End of /STATS report.", ServerNum, source);
			break;
		}
	default:
		send_cmd(NULL, "%s 219 %s %c :End of /STATS report.", ServerNum, source, *av[0]);
		break;
	}
}

/* LUSERS - forwarded */
static void m_lusers(char *source, int ac, char **av)
{
	do_lusers(source);
}

static void m_eob(char *source, int ac, char **av)
{
	if (eob_sent == 0) {
		send_cmd(NULL, "%s EA", ServerNum);
		eob_sent = 1;
	}
}

/* TIME - handled */
static char *months[] = {
	"January", "February", "March", "April",
	"May", "June", "July", "August",
	"September", "October", "November", "December"
};

static char *weekdays[] = {
	"Sunday", "Monday", "Tuesday", "Wednesday",
	"Thursday", "Friday", "Saturday"
};

static void m_time(char *source, int ac, char **av)
{
	static char buf[80], plus;
	struct tm *lt, *gm;
	struct tm gmbuf;
	int minswest;
	time_t clock;

	time(&clock);
	gm = gmtime(&clock);
	memcpy((char *) &gmbuf, (char *) gm, sizeof(gmbuf));
	gm = &gmbuf;
	lt = localtime(&clock);

	if (lt->tm_yday == gm->tm_yday)
		minswest =
			(gm->tm_hour - lt->tm_hour) * 60 + (gm->tm_min - lt->tm_min);
	else if (lt->tm_yday > gm->tm_yday)
		minswest = (gm->tm_hour - (lt->tm_hour + 24)) * 60;
	else
		minswest = ((gm->tm_hour + 24) - lt->tm_hour) * 60;

	plus = (minswest > 0) ? '-' : '+';
	if (minswest < 0)
		minswest = -minswest;

	(void) snprintf(buf, sizeof(buf),
						 "%s %s %d %04d -- %02d:%02d %c%02d:%02d",
						 weekdays[lt->tm_wday], months[lt->tm_mon], lt->tm_mday,
						 lt->tm_year + 1900, lt->tm_hour, lt->tm_min, plus,
						 minswest / 60, minswest % 60);

	send_cmd(NULL, "%s 391 %s %s :%s", ServerNum, source, ServerName, buf);
}

static void m_privmsg(char *source, int ac, char **av)
{
	int nickid;
	int chanid;
	int dbcount = 0;
	char *message = "";
	char *nick = "";
        char **res;
        MYSQL_RES *resptr;

	chanid = db_checkchannel(av[0]);
	nickid = db_getnumeric(source);

        db_query("SELECT nick FROM " TBL_USER " WHERE nickid=\'%d\'", nickid);
        resptr = mysql_store_result(myptr);
        while ((res = mysql_fetch_row(resptr)))
       	        nick = db_escape(res[0]);

	mysql_free_result(resptr);


	if (*av[0] == '#') { /* channel message */
	        db_query("SELECT count(*) FROM " TBL_CHANLOG " WHERE chanid=\'%d\'", chanid);
	        resptr = mysql_store_result(myptr);
	        while ((res = mysql_fetch_row(resptr)))
        	        dbcount = atoi(db_escape(res[0]));

		mysql_free_result(resptr);
		message = db_escape(av[1]);

		if (dbcount < LogAmount) { /* simple insert */
			db_query("INSERT INTO " TBL_CHANLOG
				 " (logid, chanid, time, nick, message) VALUES (\'%d\', \'%d\', \'%ld\', \'%s\', \'%s\')",
				 dbcount + 1, chanid, time(NULL), nick, message);
		} else { /* shove everything back by one then insert as logid 5 */
			db_query("DELETE FROM " TBL_CHANLOG " WHERE logid='1' AND chanid=\'%d\'", chanid);
			int cnt = 2;

			while (cnt < LogAmount+1) {
				db_query("UPDATE " TBL_CHANLOG
					" SET logid=\'%d\' WHERE chanid=\'%d\' AND logid=\'%d\'",
					cnt - 1, chanid, cnt);
				cnt++;
			}

			db_query("INSERT INTO " TBL_CHANLOG
				 " (logid, chanid, time, nick, message) VALUES (\'%d\', \'%d\', \'%ld\', \'%s\', \'%s\')",
				 LogAmount, chanid, time(NULL), nick, message);
		}
	} else { /* private message */
		commandcheck("P", av, ac, source);

		return; /* no use for this yet */
	}
}

/*
 * IRCD specific commands
 */

#if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
static void m_exempt(char *source, int ac, char **av)
{
	int masklen;
	int action = 0;
	char *emask;

	emask = av[1];
	masklen = strlen(emask);

	while (*emask) {
		switch (*emask) {
			case '+':
				action = 0;
				break;

			case '-':
				action = 1;
				break;

			default:
				break;
		}

		if (masklen - strlen(emask) == 1)
			break;
		emask++;
	}

#ifdef IRCD_NEFARIOUS10
	if (action == 0) {
		db_query("INSERT INTO " TBL_EXEMPT
			 " (emask, expire) VALUES (\'%s\', \'%s\')",
			 emask, av[2]);
	} else {
		db_query("DELETE FROM " TBL_EXEMPT " WHERE emask='%s'", emask);
	}
#else
	if (action == 0) {
		db_query("INSERT INTO " TBL_EXEMPT
			 " (emask) VALUES (\'%s\')",
			 emask);
	} else {
		db_query("DELETE FROM " TBL_EXEMPT " WHERE emask='%s'", emask);
	}
#endif
}

static void m_mark(char *source, int ac, char **av)
{
	int nickid;

	nickid = db_getnick(av[0]);

#ifdef IRCD_NEFARIOUS4
	/* AR MK Fizz DNSBL njabl ma njabl.adsl-68-95-146-235.dsl.rcsntx.swbell.net :http:// */
	/*       0    1     2     3  4                                              5 */
	if (!strcmp(av[1], "DNSBL")) {
		db_query("INSERT INTO " TBL_DNSBL
			 " (nickid, dnsbl, flags, host, reason) VALUES (\'%d\', \'%s\', \'%s\', \'%s\', \'%s\')",
			 nickid, av[2], av[3], av[4], av[5]);
	}
#else
	/* AS MK SiRVulcaN DNSBL      ma      TestRBL.60-234-132-220.bitstream.orcon.net.nz */
        /* AS MK SiRVulcaN DNSBL_DATA TestRBL */
	/*       0         1          2       3 */
	if (!strcmp(av[1], "DNSBL")) {
		db_query("INSERT INTO " TBL_DNSBL
			 " (nickid, flags, host) VALUES (\'%d\', \'%s\', \'%s\')",
			 nickid, av[2], av[3]);
	} else if (!strcmp(av[1], "DNSBL_DATA")) {
		int dnsblid = 0;
	        char **res;
	        MYSQL_RES *resptr;

	        db_query("SELECT dnsblid FROM " TBL_DNSBL " WHERE nickid=\'%d\'", nickid);
	        resptr = mysql_store_result(myptr);
	        while ((res = mysql_fetch_row(resptr)))
        	        dnsblid = atoi(db_escape(res[0]));

		mysql_free_result(resptr);

		db_query("INSERT INTO " TBL_DNSBL_DATA
			 " (dnsblid, nickid, dnsbl) VALUES (\'%d\', \'%d\', \'%s\')",
			 dnsblid, nickid, av[3]);
	}

#endif
}

static void m_fakehost(char *source, int ac, char **av)
{
	int nickid;

	nickid = db_getnumeric(av[0]);

	db_query("UPDATE " TBL_USER
                 " SET fakehost=\"%s\" WHERE nickid=\"%d\"",
                 av[1], nickid);
}

static void m_svsjoin(char *source, int ac, char **av)
{
	do_join(av[1], av[2], 0);
}

static void m_svspart(char *source, int ac, char **av)
{
	do_part(av[1], av[0]);
}

static void m_svsnick(char *source, int ac, char **av)
{
	do_nick_chg(av[1], av[0]);
}

static void m_swhois(char *source, int ac, char **av)
{
	do_swhois(av[0], av[1]);
}

#ifdef IRCD_NEFARIOUS10
static void m_shun(char *source, int ac, char **av)
{
	int exist = 0;
	int shunid = 0;
	int action = 0; /* 0 = add, 1 = remove */
	int masklen;
	char *lastmod;
	char *reason;
	char *smask;
        char **res;
        MYSQL_RES *resptr;

	smask = av[1];
	masklen = strlen(smask);

	while (*smask) {
		switch (*smask) {
			case '+':
				action = 0;
				break;

			case '-':
				action = 1;
				break;

			default:
				break;
		}

		if (masklen - strlen(smask) == 1)
			break;
		smask++;
	}

        db_query("SELECT shunid FROM " TBL_SHUN " WHERE smask=\'%s\'", smask);

	if (ac == 4)
		lastmod = "0";
	else
		lastmod = av[3];

        resptr = mysql_store_result(myptr);
        while ((res = mysql_fetch_row(resptr))) {
                shunid = atoi(db_escape(res[0]));
                exist = 1;
	}
	mysql_free_result(resptr);

	reason = db_escape(av[ac-1]);

	if ((exist == 0) && (action == 0))
		db_query("INSERT INTO " TBL_SHUN
			 " (smask, expire, lastmod, reason) VALUES (\'%s\', \'%ld\',\'%s\',\"%s\")",
			 smask, time(NULL) - atoi(av[2]), lastmod, reason);
	else if ((exist == 1) && (action == 0))
		db_query("UPDATE " TBL_SHUN
			 " SET smask='%s', expire='%ld', lastmod='%s', reason=\"%s\" WHERE shunid='%d'",
			 smask, time(NULL) - atoi(av[2]), lastmod, reason, shunid);
	else if ((exist == 1) && (action == 1))
		db_query("DELETE FROM " TBL_SHUN " WHERE shunid='%d'", shunid);
			
	exist = 0;
}
#endif
#endif

/*************************************************************************/
Message messages[] = {

#if defined(IRCD_IRCU11)||defined(IRCD_IRCU12)||defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)||defined(IRCD_ASUKA)

	{"PASS", NULL}
	,
	{"B", m_burst}
	,
	{"C", m_create}
	,
	{"G", m_ping}
	,
	{"SERVER", m_server}
	,
	{"S", m_server}
	,
	{"V", m_version}
	,
	{"SQ", m_squit}
	,
	{"N", m_nick}
	,
	{"D", m_kill}
	,
	{"K", m_kick}
	,
	{"L", m_part}
	,
	{"M", m_mode}
	,
	{"T", m_topic}
	,
	{"Q", m_quit}
	,
	{"R", m_stats}
	,
	{"J", m_join}
	,
	{"A", m_away}
	,
	{"LU", m_lusers}
	,
	{"TI", m_time}
	,
	{"EB", m_eob}
	,
	{"AC", m_account}
	,
	{"EA", NULL}
	,
	{"Y", m_error}
	,
	{"GL", m_gline}
	,
	{"OM", m_mode}
	,
	{"CM", m_clearmode}
	,
	{"P", m_privmsg}
	,
#endif
#if defined(IRCD_NEFARIOUS4)||defined(IRCD_NEFARIOUS10)
	/* Nefarious IRCu Numerics */
	{"EX", m_exempt}
	,
	{"FA", m_fakehost}
	,
	{"MK", m_mark}
	,
	{"SJ", m_svsjoin}
	,
	{"SN", m_svsnick}
	,
	{"SP", m_svspart}
	,
	{"SW", m_swhois}
	,
#if defined(IRCD_NEFARIOUS10)
	{"SU", m_shun }
	,
#endif
#endif
	/* ----------------------- */

	{NULL}
};

/*************************************************************************/

Message *find_message(const char *name)
{
	Message *m;
	for (m = messages; m->name; m++)
		if (stricmp(name, m->name) == 0)
			return m;
	return NULL;
}

/*************************************************************************/
